#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for the measurement calculus.

This will be used to manipulate the measurement patterns in the measurement-based quantum computation.
"""

from argparse import ArgumentTypeError
from typing import List, Any
from numpy import pi
from qcompute_qnet.quantum.pattern import Pattern
from qcompute_qnet.quantum.utils import print_progress

__all__ = [
    "MCalculus",
    "transpile_to_pattern"
]


class MCalculus:
    r"""Class for manipulating measurement patterns.

    This class provides various basic operations to compute measurement patterns.
    Please see the reference [The measurement calculus, arXiv: 0704.1263] for technical details.
    """

    def __init__(self):
        r"""Constructor for MCalculus class.
        """
        self.__circuit = None  # circuit
        self.__circuit_width = None  # circuit width
        self.__circuit_slice = []  # sliced circuit
        self.__wild_pattern = []  # wild pattern
        self.__pattern = None  # standard pattern
        self.__measured_qubits = []  # measured qubits
        self.__track = False  # progress tracking

    def set_circuit(self, circuit: "Circuit") -> None:
        r"""Set a quantum circuit.

        Args:
            circuit (Circuit): quantum circuit to set
        """
        self.__circuit = circuit
        self.__circuit_width = circuit.width

    def __slice_circuit(self, gate_history: list) -> None:
        r"""Slice a quantum circuit and label the input and output qubits for each gate.

        Args:
            gate_history (list): a list of gates in the circuit
        """
        # Slice the circuit to mark the input_/output_ labels for measurement patterns
        counter = []
        for gate in gate_history:
            name = gate["name"]
            which_qubits = gate["which_qubit"]
            qubit_number = len(which_qubits)
            if qubit_number not in [1, 2]:
                raise ArgumentTypeError(f"Invalid quantum gate: ({gate})!\n"
                                        f"({qubit_number})-qubit gate is not supported in this version.\n")

            if qubit_number == 1:  # single qubit gate
                which_qubit = which_qubits[0]  # take the qubit index
                if which_qubit in self.__measured_qubits:
                    raise ArgumentTypeError(f"Invalid qubit index: ({which_qubit})!\n"
                                            f"Please check your qubit index as this qubit has already been measured.")

                input_ = [(which_qubit, int(counter.count(which_qubit)))]

                if name == 'm':
                    # Measurements have no output labels
                    output_ = []
                else:
                    output_ = [(which_qubit, int(counter.count(which_qubit) + 1))]
                    counter += [which_qubit]  # count the qubit index

            else:  # double qubit gate
                control = which_qubits[0]  # take the control qubit
                target = which_qubits[1]  # take the target qubit
                if control in self.__measured_qubits:
                    raise ArgumentTypeError(f"Invalid qubit index: ({control})!\n"
                                            f"Please check your qubit index as this qubit has already been measured.")
                if target in self.__measured_qubits:
                    raise ArgumentTypeError(f"Invalid qubit index: ({target})!\n"
                                            f"Please check your qubit index as this qubit has already been measured.")

                input_ = [(control, int(counter.count(control))), (target, int(counter.count(target)))]

                if name == 'cz':  # input and output labels are the same
                    output_ = input_[:]

                elif name == 'cx':
                    output_ = [(control, int(counter.count(control))),
                               (target, int(counter.count(target) + 1))]
                    counter += [target]  # count the qubit index

                else:
                    output_ = [(control, int(counter.count(control) + 1)),
                               (target, int(counter.count(target) + 1))]
                    counter += which_qubits  # count the qubit index

            # The gate after slicing has a form: [original_gate, input_, output_]
            # E.g. = [[cx, [0, 1], None], input_, output_]
            self.__circuit_slice.append([gate, input_, output_])

    @staticmethod
    def __set_ancilla(input_: list, output_: list, ancilla_number_list=None) -> list:
        r"""Insert ancilla qubits.

        Insert ancilla qubits between the input and output qubits. 
        The coordinates of the ancilla qubits are evenly distributed.

        Args:
            input_ (list): input vertices
            output_ (list): output vertices
            ancilla_number_list (list): number of ancilla qubits

        Returns:
            list: ancilla qubit labels
        """
        ancilla_num_in = len(input_)
        ancilla_num_out = len(output_)
        if ancilla_num_in != ancilla_num_out:
            raise ArgumentTypeError(f"Invalid input and output labels: ({input_}) and ({output_})!\n"
                                    f"Input labels and output labels must have the same length.")
        if ancilla_num_in not in [1, 2]:
            raise ArgumentTypeError(f"Invalid ancilla number: ({ancilla_num_in})!\n"
                                    f"({ancilla_num_in})-qubit gate is not supported in this version.\n")

        ancilla_nums = [] if ancilla_number_list is None else ancilla_number_list
        ancilla_idxes = []

        for i in range(len(ancilla_nums)):
            input_qubit = input_[i]  # obtain input qubit index
            row_in = input_qubit[0]  # input qubit row
            col_in = input_qubit[1]  # input qubit column

            output_qubit = output_[i]  # obtain output qubit
            row_out = output_qubit[0]  # output qubit row
            col_out = output_qubit[1]  # output qubit column

            ancilla = ancilla_nums[i]

            if row_in != row_out:
                raise ArgumentTypeError(f"Invalid input and output labels: ({input_}) and ({output_})!\n"
                                        f"Each input label and output label must have the same qubit index.")

            # Calculate ancillary qubits' positions
            col_len = col_out - col_in
            pos = [(int(col_in * (ancilla + 1) + col * col_len) / (ancilla + 1)) for col in range(1, ancilla + 1)]

            # Set ancillary qubits' labels
            for idx in range(ancilla):
                ancilla_idxes.append((row_in, pos[idx]))

        return ancilla_idxes

    def __to_pattern(self, gate: list) -> None:
        r"""Translate a quantum gate or measurement to its equivalent pattern.

        Args:
            gate (list): gate or measurement to translate
        """
        original_gate, input_, output_ = gate
        name = original_gate["name"]
        which_qubit = original_gate["which_qubit"]

        ancilla = []

        if name == 'h':  # Hadamard gate
            E = Pattern.CommandE([input_[0], output_[0]])
            M = Pattern.CommandM(input_[0], 0, 'XY', [], [])
            X = Pattern.CommandX(output_[0], [input_[0]])
            commands = [E, M, X]

        elif name == 'x':  # Pauli X gate
            ancilla = self.__set_ancilla(input_, output_, [1])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], output_[0]])
            M1 = Pattern.CommandM(input_[0], 0, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], - pi, 'XY', [], [])
            X3 = Pattern.CommandX(output_[0], [ancilla[0]])
            Z3 = Pattern.CommandZ(output_[0], [input_[0]])
            commands = [E12, E23, M1, M2, X3, Z3]

        elif name == 'y':  # Pauli Y gate
            ancilla = self.__set_ancilla(input_, output_, [3])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], ancilla[1]])
            E34 = Pattern.CommandE([ancilla[1], ancilla[2]])
            E45 = Pattern.CommandE([ancilla[2], output_[0]])
            M1 = Pattern.CommandM(input_[0], pi / 2, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], pi / 2, 'XY', [], [])
            M3 = Pattern.CommandM(ancilla[1], - pi / 2, 'XY', [], [input_[0], ancilla[0]])
            M4 = Pattern.CommandM(ancilla[2], 0, 'XY', [], [ancilla[0]])
            X5 = Pattern.CommandX(output_[0], [ancilla[2]])
            Z5 = Pattern.CommandZ(output_[0], [ancilla[1]])
            commands = [E12, E23, E34, E45, M1, M2, M3, M4, X5, Z5]

        elif name == 'z':  # Pauli Z gate
            ancilla = self.__set_ancilla(input_, output_, [1])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], output_[0]])
            M1 = Pattern.CommandM(input_[0], - pi, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], 0, 'XY', [], [])
            X3 = Pattern.CommandX(output_[0], [ancilla[0]])
            Z3 = Pattern.CommandZ(output_[0], [input_[0]])
            commands = [E12, E23, M1, M2, X3, Z3]

        elif name == 's':  # phase gate
            ancilla = self.__set_ancilla(input_, output_, [1])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], output_[0]])
            M1 = Pattern.CommandM(input_[0], - pi / 2, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], 0, 'XY', [], [])
            X3 = Pattern.CommandX(output_[0], [ancilla[0]])
            Z3 = Pattern.CommandZ(output_[0], [input_[0]])
            commands = [E12, E23, M1, M2, X3, Z3]

        elif name == 't':  # T gate
            ancilla = self.__set_ancilla(input_, output_, [1])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], output_[0]])
            M1 = Pattern.CommandM(input_[0], - pi / 4, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], 0, 'XY', [], [])
            X3 = Pattern.CommandX(output_[0], [ancilla[0]])
            Z3 = Pattern.CommandZ(output_[0], [input_[0]])
            commands = [E12, E23, M1, M2, X3, Z3]

        elif name == 'rx':  # rotation gate around x-axis
            ancilla = self.__set_ancilla(input_, output_, [1])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], output_[0]])
            M1 = Pattern.CommandM(input_[0], 0, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], - original_gate["angle"], 'XY', [input_[0]], [])
            X3 = Pattern.CommandX(output_[0], [ancilla[0]])
            Z3 = Pattern.CommandZ(output_[0], [input_[0]])
            commands = [E12, E23, M1, M2, X3, Z3]

        elif name == 'ry':  # rotation gate around y-axis
            ancilla = self.__set_ancilla(input_, output_, [3])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], ancilla[1]])
            E34 = Pattern.CommandE([ancilla[1], ancilla[2]])
            E45 = Pattern.CommandE([ancilla[2], output_[0]])
            M1 = Pattern.CommandM(input_[0], pi / 2, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], - original_gate["angle"], 'XY', [input_[0]], [])
            M3 = Pattern.CommandM(ancilla[1], - pi / 2, 'XY', [], [input_[0], ancilla[0]])
            M4 = Pattern.CommandM(ancilla[2], 0, 'XY', [], [ancilla[0]])
            X5 = Pattern.CommandX(output_[0], [ancilla[2]])
            Z5 = Pattern.CommandZ(output_[0], [ancilla[1]])
            commands = [E12, E23, E34, E45, M1, M2, M3, M4, X5, Z5]

        elif name == 'rz':  # rotation gate around z axis
            ancilla = self.__set_ancilla(input_, output_, [1])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], output_[0]])
            M1 = Pattern.CommandM(input_[0], - original_gate["angle"], 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], 0, 'XY', [], [])
            X3 = Pattern.CommandX(output_[0], [ancilla[0]])
            Z3 = Pattern.CommandZ(output_[0], [input_[0]])
            commands = [E12, E23, M1, M2, X3, Z3]

        elif name == 'rz5':  # rotation gate around z axis
            ancilla = self.__set_ancilla(input_, output_, [3])
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], ancilla[1]])
            E34 = Pattern.CommandE([ancilla[1], ancilla[2]])
            E45 = Pattern.CommandE([ancilla[2], output_[0]])
            M1 = Pattern.CommandM(input_[0], 0, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], 0, 'XY', [], [])
            M3 = Pattern.CommandM(ancilla[1], - original_gate["angle"], 'XY', [ancilla[0]], [input_[0]])
            M4 = Pattern.CommandM(ancilla[2], 0, 'XY', [], [ancilla[0]])
            X5 = Pattern.CommandX(output_[0], [ancilla[2]])
            Z5 = Pattern.CommandZ(output_[0], [ancilla[1]])
            commands = [E12, E23, E34, E45, M1, M2, M3, M4, X5, Z5]

        elif name == 'u3':  # single-qubit unitary gate
            # Note: U = Rz(\phi) Rx(\theta) Rz(\lamda)
            ancilla = self.__set_ancilla(input_, output_, [3])
            theta, phi, gamma = original_gate["angles"]
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], ancilla[1]])
            E34 = Pattern.CommandE([ancilla[1], ancilla[2]])
            E45 = Pattern.CommandE([ancilla[2], output_[0]])
            M1 = Pattern.CommandM(input_[0], - gamma, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], - theta, 'XY', [input_[0]], [])
            M3 = Pattern.CommandM(ancilla[1], - phi, 'XY', [ancilla[0]], [input_[0]])
            M4 = Pattern.CommandM(ancilla[2], 0, 'XY', [], [ancilla[0]])
            X5 = Pattern.CommandX(output_[0], [ancilla[2]])
            Z5 = Pattern.CommandZ(output_[0], [ancilla[1]])
            commands = [E12, E23, E34, E45, M1, M2, M3, M4, X5, Z5]

        elif name == 'u':  # single-qubit unitary gate
            # Note: U = Rz(\phi) Ry(\theta) Rz(\lamda) usually used in MBQC
            ancilla = self.__set_ancilla(input_, output_, [3])
            theta, phi, gamma = original_gate["angles"]
            E12 = Pattern.CommandE([input_[0], ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], ancilla[1]])
            E34 = Pattern.CommandE([ancilla[1], ancilla[2]])
            E45 = Pattern.CommandE([ancilla[2], output_[0]])
            M1 = Pattern.CommandM(input_[0], - (gamma + - pi / 2), 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], - theta, 'XY', [input_[0]], [])
            M3 = Pattern.CommandM(ancilla[1], - (phi + pi / 2), 'XY', [ancilla[0]], [input_[0]])
            M4 = Pattern.CommandM(ancilla[2], 0, 'XY', [], [ancilla[0]])
            X5 = Pattern.CommandX(output_[0], [ancilla[2]])
            Z5 = Pattern.CommandZ(output_[0], [ancilla[1]])
            commands = [E12, E23, E34, E45, M1, M2, M3, M4, X5, Z5]

        elif name == 'cx':  # control NOT gate
            ancilla = self.__set_ancilla(input_, output_, [0, 1])
            E23 = Pattern.CommandE([input_[1], ancilla[0]])
            E13 = Pattern.CommandE([input_[0], ancilla[0]])
            E34 = Pattern.CommandE([ancilla[0], output_[1]])
            M2 = Pattern.CommandM(input_[1], 0, 'XY', [], [])
            M3 = Pattern.CommandM(ancilla[0], 0, 'XY', [], [])
            X4 = Pattern.CommandX(output_[1], [ancilla[0]])
            Z1 = Pattern.CommandZ(output_[0], [input_[1]])
            Z4 = Pattern.CommandZ(output_[1], [input_[1]])
            commands = [E23, E13, E34, M2, M3, X4, Z1, Z4]

        # Measurement pattern of CNOT with 15 qubits, c.f. [arXiv: quant-ph/0301052v2]
        # Note: due to the '1' in byproduct Z of qubit 7, we manually add a Z gate after qubit 7 to eliminate this
        elif name == 'cnot15':  # control not gate
            input1, input2 = input_
            output1, output2 = output_
            ancilla = self.__set_ancilla(input_, output_, [7, 5])
            new_row = (input2[0] + input1[0]) / 2
            new_col_1 = output1[1] + input1[1]
            new_col_2 = output2[1] + input2[1]
            new_col = (new_col_1 + new_col_2) / 4
            ancilla.append((new_row, new_col))
            E12 = Pattern.CommandE([input1, ancilla[0]])
            E23 = Pattern.CommandE([ancilla[0], ancilla[1]])
            E34 = Pattern.CommandE([ancilla[1], ancilla[2]])
            E45 = Pattern.CommandE([ancilla[2], ancilla[3]])
            E48 = Pattern.CommandE([ancilla[2], ancilla[12]])
            E56 = Pattern.CommandE([ancilla[3], ancilla[4]])
            E67 = Pattern.CommandE([ancilla[4], ancilla[5]])
            E716 = Pattern.CommandE([ancilla[5], ancilla[6]])
            E1617 = Pattern.CommandE([ancilla[6], output1])
            E910 = Pattern.CommandE([input2, ancilla[7]])
            E1011 = Pattern.CommandE([ancilla[7], ancilla[8]])
            E1112 = Pattern.CommandE([ancilla[8], ancilla[9]])
            E812 = Pattern.CommandE([ancilla[12], ancilla[9]])
            E1213 = Pattern.CommandE([ancilla[9], ancilla[10]])
            E1314 = Pattern.CommandE([ancilla[10], ancilla[11]])
            E1415 = Pattern.CommandE([ancilla[11], output2])
            M1 = Pattern.CommandM(input1, 0, 'XY', [], [])
            M2 = Pattern.CommandM(ancilla[0], pi / 2, 'XY', [], [])
            M3 = Pattern.CommandM(ancilla[1], pi / 2, 'XY', [], [])
            M4 = Pattern.CommandM(ancilla[2], pi / 2, 'XY', [], [])
            M5 = Pattern.CommandM(ancilla[3], pi / 2, 'XY', [], [])
            M6 = Pattern.CommandM(ancilla[4], pi / 2, 'XY', [], [])
            M8 = Pattern.CommandM(ancilla[12], pi / 2, 'XY', [], [])
            M12 = Pattern.CommandM(ancilla[9], pi / 2, 'XY', [], [])
            M9 = Pattern.CommandM(input2, 0, 'XY', [], [])
            M10 = Pattern.CommandM(ancilla[7], 0, 'XY', [], [])
            M11 = Pattern.CommandM(ancilla[8], 0, 'XY', [], [])
            M13 = Pattern.CommandM(ancilla[10], 0, 'XY', [], [])
            M14 = Pattern.CommandM(ancilla[11], 0, 'XY', [], [])
            M7 = Pattern.CommandM(ancilla[5], - pi, 'XY', [], input_ + [ancilla[i] for i in [1, 2, 3, 8, 12]])
            M16 = Pattern.CommandM(ancilla[6], 0, 'XY', [], [ancilla[i] for i in [0, 1, 3, 4]])
            X15 = Pattern.CommandX(output2, [ancilla[i] for i in [0, 1, 7, 9, 11, 12]])
            X17 = Pattern.CommandX(output1, [ancilla[6]])
            Z15 = Pattern.CommandZ(output2, [input2, ancilla[8], ancilla[10]])
            Z17 = Pattern.CommandZ(output1, [ancilla[5]])
            commands = [E12, E23, E34, E45, E48, E56, E67, E716, E1617, E910, E1011, E1112, E812, E1213, E1314, E1415,
                        M1, M2, M3, M4, M5, M6, M8, M12, M9, M10, M11, M13, M14, M7, M16, X15, X17, Z15, Z17]

        elif name == 'cz':  # control Z gate
            commands = [Pattern.CommandE(input_)]

        elif name == 'm':  # single qubit measurement
            self.__measured_qubits.append(which_qubit[0])
            commands = [Pattern.CommandM(input_[0], *original_gate["basis"])]

        else:
            raise ArgumentTypeError(f"Invalid gate: ({gate})!\n"
                                    f"Translation of such a gate is not supported in this version.")

        self.__wild_pattern.append(
            Pattern(str([name] + [str(qubit) for qubit in which_qubit]), list(set(input_ + output_ + ancilla)),
                    input_, output_, commands))

    def __list_to_net(self, vertices: list) -> List[List]:
        r"""Map a vertex list to a vertex net.

        This method maps a one-dimensional vertex list to a two-dimensional vertex net.
        The net contains the labels of all vertices on the graph.

        Args:
            vertices (list): one-dimensional list containing the labels of input or output vertices

        Returns:
            list: a two-dimensional vertex net. This net has a special network structure of List[List],
                  Each row of the net records the column number of input or output vertices.
        """
        # Get width
        rows = {qubit[0] for qubit in vertices}
        self.__circuit_width = len(rows) if self.__circuit_width is None else self.__circuit_width
        # Map to net
        bit_net = [[] for _ in range(self.__circuit_width)]
        for qubit in vertices:
            bit_net[qubit[0]].append(qubit[1])
        return bit_net

    def __get_input(self) -> List[Any]:
        r"""Get the labels of the input vertices in the list.

        Returns:
            list: a list containing the labels of input vertices
        """
        input_list = []
        for pat in self.__wild_pattern:
            input_list += pat.input_
        # Map to net
        bit_net = self.__list_to_net(input_list)
        input_ = [(i, min(bit_net[i])) for i in range(self.__circuit_width)]
        return input_

    def __get_output(self) -> List[Any]:
        r"""Get the labels of the output vertices in the list.

        Returns:
            list: a list containing the labels of output vertices
        """
        output_list = []
        for pat in self.__wild_pattern:
            output_list += pat.output_
        bit_net = self.__list_to_net(output_list)
        # Divide the output into classical channels and quantum channels
        c_out = [(i, max(bit_net[i])) for i in range(self.__circuit_width) if i in self.__measured_qubits]
        q_out = [(i, max(bit_net[i])) for i in range(self.__circuit_width) if i not in self.__measured_qubits]
        output_ = [c_out, q_out]
        return output_

    def __splice_patterns(self, input_: list, output_: list) -> None:
        r"""Splice all measurement patterns together to form a complete pattern.

        The spliced pattern is equivalent to the complete quantum circuit.

        Args:
            input_ (list): pattern input vertices
            output_ (list): pattern output vertices
        """
        names = ''
        space = []
        cmds = []

        for pat in self.__wild_pattern:
            names += pat.name
            space += [qubit for qubit in pat.space if qubit not in space]
            cmds += pat.commands

        self.__pattern = Pattern(names, space, input_, output_, cmds)

    def to_pattern(self) -> None:
        r"""Translate a quantum circuit to its equivalent measurement pattern.
        """
        self.__slice_circuit(self.__circuit.gate_history)
        for gate in self.__circuit_slice:
            self.__to_pattern(gate)

        input_ = self.__get_input()
        output_ = self.__get_output()

        self.__splice_patterns(input_, output_)

    @staticmethod
    def __swap(two_cmds: list) -> list:
        r"""Swap two commands.

        Hint:
            The commands in the list are executed from the left to the right as default.
            For example, the command 'E' represents the entanglement command.
            According to the MBQC procedure, entanglements are executed in the first step.
            Therefore, the command 'E' should be on the left side of the list.
            After the entanglements, the measurement commands are executed, and so on.

        Warning:
            This method only swaps two commands in ['E', 'M', 'X', 'Z', 'S'].
            Swapping two commands is non-trivial. Please refer to [arXiv:0704.1263] for technical details.

        Args:
            two_cmds (list): a list of two commands to swap

        Returns:
            list: a list of two new commands after swapping
        """
        cmd1 = two_cmds[0]
        cmd2 = two_cmds[1]
        name1 = cmd1.name
        name2 = cmd2.name

        if not {name1, name2}.issubset(['E', 'M', 'X', 'Z', 'S']):
            raise ArgumentTypeError(f"Invalid command names: ({name1}) and ({name2})!"
                                    "Only 'E', 'M', 'X', 'Z' and 'S' are supported as the command names.")

        # [X, E] --> [E, X]
        if name1 == 'X' and name2 == 'E':
            X_qubit = cmd1.which_qubit
            E_qubits = cmd2.which_qubits[:]
            if X_qubit not in E_qubits:
                return [cmd2, cmd1]  # swap the independent commands
            else:
                op_qubit = list(set(E_qubits).difference([X_qubit]))
                new_cmd = Pattern.CommandZ(op_qubit[0], cmd1.domain)
                return [cmd2, new_cmd, cmd1]  # create a new command

        # [Z, E] --> [E, Z]
        elif name1 == 'Z' and name2 == 'E':
            return [cmd2, cmd1]  # swap the independent commands

        # [M, E] --> [E, M]
        elif name1 == 'M' and name2 == 'E':
            if cmd1.which_qubit not in cmd2.which_qubits:
                return [cmd2, cmd1]  # swap the independent commands
            else:
                raise ArgumentTypeError(f"Invalid command names: ({name1}) and ({name2})!"
                                        f"The measurement command must be executed after the entanglement command.")

        # [X, M] --> [M, X]
        elif name1 == 'X' and name2 == 'M':
            X_qubit = cmd1.which_qubit
            M_qubit = cmd2.which_qubit
            if X_qubit != M_qubit:
                return [cmd2, cmd1]  # swap the independent commands
            else:
                measurement_plane = cmd2.plane
                if measurement_plane == 'XY':
                    M_new = Pattern.CommandM(M_qubit, cmd2.angle, 'XY', cmd2.domain_s + cmd1.domain, cmd2.domain_t)
                elif measurement_plane == 'YZ':
                    M_new = Pattern.CommandM(M_qubit, cmd2.angle, 'YZ', cmd2.domain_s, cmd2.domain_t + cmd1.domain)
                else:
                    raise ArgumentTypeError(f"Invalid measurement plane: ({measurement_plane})! Only 'XY' and 'YZ' are"
                                            f" supported as the measurement plane in this version.")

                return [M_new]  # create a new command

        # [Z, M] --> [M, Z]
        elif name1 == 'Z' and name2 == 'M':
            Z_qubit = cmd1.which_qubit
            M_qubit = cmd2.which_qubit
            if Z_qubit != M_qubit:
                return [cmd2, cmd1]  # swap the independent commands
            else:
                measurement_plane = cmd2.plane
                if measurement_plane == 'YZ':
                    M_new = Pattern.CommandM(M_qubit, cmd2.angle, 'YZ', cmd2.domain_s + cmd1.domain, cmd2.domain_t)
                elif measurement_plane == 'XY':
                    M_new = Pattern.CommandM(M_qubit, cmd2.angle, 'XY', cmd2.domain_s, cmd2.domain_t + cmd1.domain)
                else:
                    raise ArgumentTypeError(f"Invalid measurement plane: ({measurement_plane})! Only 'XY' and 'YZ' are"
                                            f" supported as the measurement plane in this version.")
                return [M_new]  # create a new command

        # [Z, X] --> [X, Z]
        elif name1 == 'Z' and name2 == 'X':
            return [cmd2, cmd1]  # swap the independent commands

        # [X, X] --> [X]
        elif name1 == 'X' and name2 == 'X':
            X1_qubit = cmd1.which_qubit
            X2_qubit = cmd2.which_qubit
            if X1_qubit == X2_qubit:
                return [Pattern.CommandX(X1_qubit, cmd1.domain + cmd2.domain)]  # create a new command
            else:
                return two_cmds  # keep both commands

        # [Z, Z] --> [Z]
        elif name1 == 'Z' and name2 == 'Z':
            Z1_qubit = cmd1.which_qubit
            Z2_qubit = cmd2.which_qubit
            if Z1_qubit == Z2_qubit:
                return [Pattern.CommandZ(Z1_qubit, cmd1.domain + cmd2.domain)]  # create a new command
            else:
                return two_cmds  # keep both commands

        # [S, M / X / Z / S] --> [M / X / Z / S, S]
        elif name1 == 'S':
            # Swap command S with command M
            # Measurements in 'XY' plane and in 'YZ' plane have the same rule
            # [S, M] --> [M, S]
            if name2 == 'M':
                S_qubit = cmd1.which_qubit
                # According to the reference [arXiv:0704.1263],
                # if this 'S_qubit' is in the 'domain_s' or 'domain_t' of the command M,
                # the 'domain_s' or 'domain_t' will be summed with the domain of command S
                S_domains = cmd1.domain
                if S_qubit in cmd2.domain_s:
                    cmd2.domain_s += S_domains
                if S_qubit in cmd2.domain_t:
                    cmd2.domain_t += S_domains
                # According to the reference [arXiv:0704.1263],
                # if this 'S_qubit' is not in the 'domain_s' or 'domain_t' of the command M,
                # command S and command M can swap independently without domain modification
                return [cmd2, cmd1]  # swap the commands with domain modification
            # [S, X] --> [X, S], [S, Z] --> [Z, S], [S, S] --> [S, S]
            elif name2 == 'X' or name2 == 'Z' or name2 == 'S':
                S_qubit = cmd1.which_qubit
                S_domains = cmd1.domain
                if S_qubit in cmd2.domain:
                    cmd2.domain += S_domains
                return [cmd2, cmd1]  # swap the commands with domain modification
            else:
                return two_cmds  # keep both commands

        else:
            return two_cmds  # keep both commands

    def __propagate(self, cmd_type: str, cmds: list) -> list:
        r"""Propagate a given type of commands in the list to the front.

        Warning:
            Propagation to the front does not mean to swap this type of commands
            with brute force to the leftmost of the whole list.
            Instead, just propagate this type of commands to the available leftmost positions.
            For example, the measurement commands must be executed after the entanglement commands.
            So when calling this method, all measurement commands are propagated to the
            leftmost positions after the entanglement commands.

        Args:
            cmd_type (str): a command type to be propagated. It must be in ["E", "M", "X", "Z", "S"]
            cmds (list): a list of commands

        Returns:
            list: a list of new commands after propagation
        """
        if not isinstance(cmds, list):
            raise ArgumentTypeError(f"Invalid commands list ({cmds}) with the type: ({type(cmds)})!\n"
                                    f"Only `List` is supported as the type of commands list.")

        # Back to front, propagate command E, command M, command X and command Z sequentially
        if {cmd_type}.issubset(['E', 'M', 'X', 'Z']):
            for i in range(len(cmds) - 1, 0, -1):
                if cmds[i].name == cmd_type:
                    cmds = cmds[:i - 1] + self.__swap([cmds[i - 1], cmds[i]]) + cmds[i + 1:]

        # Front to back, propagate command S
        else:
            for i in range(0, len(cmds) - 1):
                if cmds[i].name == cmd_type:
                    cmds = cmds[:i] + self.__swap([cmds[i], cmds[i + 1]]) + cmds[i + 2:]

        return cmds

    @staticmethod
    def __cmds_to_nums(cmds: list):
        r"""Map the commands to the numbers.

        Args:
            cmds (list): a list of commands

        Returns:
            list: a list of counts on each type of the commands
            list: a list of numbers
            list: a list of numbers arranged from small to large
        """
        cmd_map = {'E': 1,  # order of commands E
                   'M': 2,  # order of commands M
                   'X': 3,  # order of commands X
                   'Z': 4,  # order of commands Z
                   'S': 5}  # order of commands S
        cmd_num_wild = [cmd_map[cmd.name] for cmd in cmds]
        cmd_num_std = cmd_num_wild[:]
        cmd_num_std.sort(reverse=False)
        cmds_count = [cmd_num_std.count(i) for i in [1, 2, 3, 4, 5]]  # count each type of commands

        return cmds_count, cmd_num_wild, cmd_num_std

    def __distance_to_standard(self, cmds: list) -> float:
        r"""Hamming distance of the current commands to the standard order.

        Args:
            cmds (list): current command list

        Returns:
            float: the hamming distance divided by the total length
        """
        cmds_count, cmd_wild, cmd_std = self.__cmds_to_nums(cmds[:])

        return sum([cmd_wild[i] == cmd_std[i] for i in range(len(cmd_wild))]) / len(cmd_wild)

    def __is_standard(self, cmd_type: str, cmds: list) -> bool:
        r"""Check if a certain type of commands is in a standard order.

        Args:
            cmd_type (str): a command type in ["E", "M", "X", "Z", "S"]
            cmds (list): a list of commands

        Returns:
            bool: whether the commands are in a standard order
        """
        if not {cmd_type}.issubset(['E', 'M', 'X', 'Z', 'S']):
            raise ArgumentTypeError(f"Input {cmd_type} should be 'E', 'M', 'X', 'Z' or 'S'.")

        # Map the commands to numbers
        cmds_count, cmd_num_wild, cmd_num_std = self.__cmds_to_nums(cmds)
        pointer_map = {'E': sum(cmds_count[:1]),  # number of commands E
                       'M': sum(cmds_count[:2]),  # number of commands E + M
                       'X': sum(cmds_count[:3]),  # number of commands E + M + X
                       'Z': sum(cmds_count[:4]),  # number of commands E + M + X + Z
                       'S': sum(cmds_count[:5])}  # number of commands E + M + X + Z + S

        return cmd_num_wild[:pointer_map[cmd_type]] == cmd_num_std[:pointer_map[cmd_type]]

    def __simplify_pauli_measurements(self):
        r"""Simplify the dependencies of Pauli measurements.

        In certain cases, the dependencies to other vertices can be excluded with Pauli simplification method.

         .. math::

            \text{Let } \alpha \text{ be a measurement angle without adaptation. }
            \text{Indeed, the adaptive angle of } \alpha \text{ has only these four values:}

            \theta_{\text{ad}} = \alpha

            \theta_{\text{ad}} = \alpha + \pi

            \theta_{\text{ad}} = - \alpha

            \theta_{\text{ad}} = - \alpha + \pi

            \text{When the } \alpha \text{ is in } [0, \pi / 2, \pi, 3 \times \pi / 2]
            \text{, the dependencies can be simplified. For example, when } \alpha = \pi
            \text{, the measurement angle of } \pi + t \times \pi \text{ and } - \pi + t \times \pi
            \text{ are equal. Therefore, it has no dependencies to the vertices in } domain\_s
            \text{. So the } domain\_s \text{ can be excluded, and so on.}
        """
        for cmd in self.__pattern.commands:
            if cmd.name == 'M':
                remainder = cmd.angle % (2 * pi)
                if remainder in [0, pi]:
                    cmd.domain_s = []
                elif remainder in [pi / 2, (3 * pi) / 2]:
                    cmd.domain_t += cmd.domain_s[:]
                    cmd.domain_s = []

    def standardize(self):
        r"""Standardize the measurement pattern.

        This method transforms a wild pattern to a standard 'EMC' pattern.
        Entanglement commands are propagated to the leftmost of the command list.
        Then come to the measurement commands and last come to the byproduct correction commands.
        To simplify the pattern, Pauli simplification method is automatically implemented to
        exclude some dependencies.
        """
        cmds = self.__pattern.commands

        for cmd_type in ['E', 'M', 'X', 'Z']:
            while not self.__is_standard(cmd_type, cmds):
                cmds = self.__propagate(cmd_type, cmds)
                print_progress(self.__distance_to_standard(cmds), "Standardization Progress", self.__track)

        self.__pattern.commands = cmds
        self.__simplify_pauli_measurements()

    @staticmethod
    def __pull_out_domain_t(cmds: list) -> list:
        r"""Pull out the signal shifting from the measurement command.

        Args:
            cmds (list): command list

        Returns:
            list: processed command list
        """
        cmds_len = len(cmds)
        for i in range(cmds_len - 1, -1, -1):
            cmd = cmds[i]
            if cmd.name == 'M':
                signal_cmd = Pattern.CommandS(cmd.which_qubit, cmd.domain_t)
                cmd.domain_t = []
                cmds = cmds[:i] + [cmd, signal_cmd] + cmds[i + 1:]
        return cmds

    def shift_signals(self) -> None:
        r"""Shift signal commands.
        """
        cmds = self.__pattern.commands
        cmds = self.__pull_out_domain_t(cmds)

        # Propagate CommandS
        while not self.__is_standard('S', cmds):
            cmds = self.__propagate('S', cmds)
            print_progress(self.__distance_to_standard(cmds), "Signal Shifting Progress", self.__track)

        self.__pattern.commands = cmds

    def get_pattern(self) -> "Pattern":
        r"""Get the measurement pattern.

        Returns:
            Pattern: a measurement pattern
        """
        return self.__pattern

    def track_progress(self, track=True):
        r"""Track the progress of measurement calculus.

        Args:
            track (bool, optional): whether to turn on the progress tracking
        """
        self.__track = track

    @staticmethod
    def __default_order_by_row(labels: list) -> list:
        r"""The default order of labels by row.

        Ordering rule: smaller row/column number has a higher priority.

        Args:
            labels (list): a list of labels

        Returns:
            list: a list of ordered labels
        """
        # Construct a dict by string labels and their float values
        labels_dict = {label: (label[0], label[1]) for label in labels}
        # Sort the dict by values (sort row first and then column)
        sorted_dict = dict(sorted(labels_dict.items(), key=lambda item: item[1]))
        # Extract the keys in the dict
        labels_sorted = list(sorted_dict.keys())

        return labels_sorted

    def optimize_by_row(self):
        r"""Optimize the measurement orders by the row-first principle.

        Warning:
            This is a heuristic algorithm that aims to measure each qubits row by row.
        """
        cmds = self.__pattern.commands

        # Split the commands by type
        cmdE_list = [cmd for cmd in cmds if cmd.name == 'E']
        cmdM_list = [cmd for cmd in cmds if cmd.name == 'M']
        cmdC_list = [cmd for cmd in cmds if cmd.name in ['X', 'Z']]
        cmdS_list = [cmd for cmd in cmds if cmd.name == 'S']

        # Construct a dict from qubit labels and their measurement commands
        cmdM_map = {cmd.which_qubit: cmd for cmd in cmdM_list}

        # Sort all the qubit labels by row
        cmdM_qubit_list = self.__default_order_by_row([cmdM.which_qubit for cmdM in cmdM_list])
        mea_length = len(cmdM_qubit_list)

        for i in range(mea_length):
            optimal = False
            while not optimal:  # if the qubits list is not standard
                # Slice measurement qubit list into three parts
                measured = cmdM_qubit_list[:i]
                measuring = cmdM_qubit_list[i]
                to_measure = cmdM_qubit_list[i + 1:]

                domains = set(cmdM_map[measuring].domain_s + cmdM_map[measuring].domain_t)
                # Find the qubits in domain but not in front of the current measurement
                push = self.__default_order_by_row(list(domains.difference(measured)))
                if push:  # Remove qubits from the to_measure list and push it to the front
                    to_measure = [qubit for qubit in to_measure if qubit not in push]
                    cmdM_qubit_list = cmdM_qubit_list[:i] + push + [measuring] + to_measure
                else:  # if no push qubits then jump out of the while loop
                    optimal = True
            print_progress((i + 1) / mea_length, "Optimization Progress", self.__track)

        # Sort the measurement commands by the sorted qubit labels
        cmdM_opt = [cmdM_map[which_qubit] for which_qubit in cmdM_qubit_list]

        # Update pattern
        cmds = cmdE_list + cmdM_opt + cmdC_list + cmdS_list
        self.__pattern.commands = cmds


def transpile_to_pattern(circuit: "Circuit", shift_signal=False, optimize=True, track=False) -> "Pattern":
    r"""Transpile a quantum circuit to the equivalent measurement pattern.

    Args:
        circuit (Circuit): quantum circuit to transpile
        shift_signal (bool): whether to perform signal sifting after standardization
        optimize (bool): whether to optimize the measurement orders
        track (bool): whether to track the transpiling progress

    Warnings:
        We should check if the circuit has sequential registers first.
        If not, we need to perform remapping before transpiling the circuit.

    Returns:
        Pattern: an equivalent measurement pattern
    """
    # Check if circuit has sequential registers
    if circuit.width != max(circuit.occupied_indices) + 1:
        new_circuit = circuit.copy()
        new_circuit.remap_indices()
        circuit = new_circuit

    mc = MCalculus()
    mc.track_progress(track)
    mc.set_circuit(circuit)
    mc.to_pattern()
    mc.standardize()
    if shift_signal:
        mc.shift_signals()
    if optimize:
        mc.optimize_by_row()
    pattern = mc.get_pattern()
    return pattern



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for quantum circuits.
"""

import copy
from argparse import ArgumentTypeError
from enum import Enum
from typing import List, Optional, Union
import numpy as np
import pandas as pd
from qcompute_qnet.quantum.backends import qcompute, Backend, mbqc

__all__ = [
    "Circuit"
]


class Circuit:
    r"""Class for creating a quantum circuit.

    Warning:
        The current version only supports gates in [I, H, X, Y, Z, S, T, Rx, Ry, Rz, U3,
        CH, CX / CNOT, CY, CZ, CRx, CRy, CRz, CU3, SWAP].

    Attributes:
        name (str): name of the circuit
        agenda (list): agenda of the circuit
        _history (List[dict]): history of the quantum gates
    """

    def __init__(self, name=None):
        r"""Constructor for Circuit class.
        """
        self.name = 'Circuit' if name is None else name
        self.agenda = []
        self.__num_qreg_unit = -1
        self._history = []

    def init_new_qreg_unit(self) -> int:
        r"""Initialize a new quantum register unit.

        Returns:
            int: address of the new quantum register unit
        """
        self.__num_qreg_unit += 1
        return self.__num_qreg_unit

    @property
    def occupied_indices(self) -> List[int]:
        r"""Get the occupied register indices in the current circuit.

        Returns:
            List[int]: occupied register indices
        """
        occupied_indices = []
        for gate in self._history:
            occupied_indices += gate["which_qubit"]

        return list(set(occupied_indices))

    @property
    def width(self) -> int:
        r"""Return the quantum circuit width.

        Returns:
            int: circuit width
        """
        return len(self.occupied_indices)

    @property
    def measured_qubits(self) -> List[int]:
        r"""Get the measured qubits in the current circuit.

        Returns:
            List[int]: measured qubits
        """
        measured_qubits = []
        for gate in self._history:
            if gate["name"] == 'm':
                measured_qubits += gate["which_qubit"]

        return list(set(measured_qubits))

    @property
    def gate_history(self) -> List[dict]:
        r"""Return the gate history of the quantum circuit.

        Returns:
            List[dict]: a list of quantum gates
        """
        return self._history

    @staticmethod
    def __check_rotation_angle(angle: Union[float, int]) -> None:
        r"""Check format of the rotation angle.

        Args:
            angle (Union[float, int]): rotation angle to check
        """
        assert isinstance(angle, float) or isinstance(angle, int), \
            f"Invalid rotation angle {angle.__repr__()} with {type(angle)} type! " \
            "Only 'float' and 'int' are supported as the type of rotation angle."

    def __check_qubit_validity(self, qubit: int) -> None:
        r"""Check validity of the qubit.

        Args:
            qubit (int): qubit to check validity
        """
        assert isinstance(qubit, int), f"Invalid qubit index {qubit.__repr__()} with {type(qubit)} type! " \
                                       f"Only 'int' is supported as the type of qubit index."
        assert qubit not in self.measured_qubits, f"Invalid qubit index: {qubit}! " \
                                                  "This qubit has already been measured."

    def __add_single_qubit_gate(self, name: str, which_qubit: int, signature=None, **params) -> None:
        r"""Add a single qubit gate to the circuit list.

        Args:
            name (str): single qubit gate name
            which_qubit (int): qubit index
            signature (Any): signature of the operation
            **params (Any): gate parameters
        """
        self.__check_qubit_validity(which_qubit)

        gate = {"name": name, "which_qubit": [which_qubit], "signature": signature, **params}
        self._history.append(gate)

    def __add_double_qubit_gate(self, name: str, which_qubit: List[int], signature=None, **params) -> None:
        r"""Add a double qubit gate to the circuit list.

        Args:
            name (str): double qubit gate name
            which_qubit (list): qubit indices in the order of [control, target]
            signature (Any): signature of the operation
            **params (Any): gate parameters
        """
        ctrl, targ = which_qubit

        self.__check_qubit_validity(ctrl)
        self.__check_qubit_validity(targ)
        if ctrl == targ:
            raise TypeError(f"Invalid qubit indices: {ctrl} and {targ}!\n"
                            "Control qubit must not be the same as target qubit.")

        gate = {"name": name, "which_qubit": which_qubit, "signature": signature, **params}
        self._history.append(gate)

    def id(self, which_qubit: int, signature=None) -> None:
        r"""Add an identity gate.

        The matrix form is:

        .. math::

            I = \begin{bmatrix} 1 & 0 \\ 0 & 1 \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the identity gate
        """
        self.__add_single_qubit_gate('id', which_qubit, signature)

    def h(self, which_qubit: int, signature=None, **condition) -> None:
        r"""Add a Hadamard gate.

        The matrix form is:

        .. math::

            H = \frac{1}{\sqrt{2}} \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the Hadamard gate
            **condition (int): condition of the operation
        """
        self.__add_single_qubit_gate('h', which_qubit, signature, **condition)

    def x(self, which_qubit: int, signature=None, **condition) -> None:
        r"""Add a Pauli-X gate.

        The matrix form is:

        .. math::

            X = \begin{bmatrix} 0 & 1 \\ 1 & 0 \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the Pauli-X gate
            **condition (int): condition of the operation
        """
        self.__add_single_qubit_gate('x', which_qubit, signature, **condition)

    def y(self, which_qubit: int, signature=None, **condition) -> None:
        r"""Add a Pauli-Y gate.

        The matrix form is:

        .. math::

            Y = \begin{bmatrix} 0 & -i \\ i & 0 \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the Pauli-Y gate
            **condition (int): condition of the operation
        """
        self.__add_single_qubit_gate('y', which_qubit, signature, **condition)

    def z(self, which_qubit: int, signature=None, **condition) -> None:
        r"""Add a Pauli-Z gate.

        The matrix form is:

        .. math::

            Z = \begin{bmatrix} 1 & 0 \\ 0 & -1 \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the Pauli-Z gate
            **condition (int): condition of the operation
        """
        self.__add_single_qubit_gate('z', which_qubit, signature, **condition)

    def s(self, which_qubit: int, signature=None) -> None:
        r"""Add a S gate.

        The matrix form is:

        .. math::

            S = \begin{bmatrix} 1 & 0 \\ 0 & i \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the S gate
        """
        self.__add_single_qubit_gate('s', which_qubit, signature)

    def t(self, which_qubit: int, signature=None) -> None:
        r"""Add a T gate.

        The matrix form is:

        .. math::

            T = \begin{bmatrix} 1 & 0 \\ 0 & e^{i \pi / 4} \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the T gate
        """
        self.__add_single_qubit_gate('t', which_qubit, signature)

    def rx(self, which_qubit: int, theta: Union[float, int], signature=None, **condition) -> None:
        r"""Add a rotation gate around x-axis.

        The matrix form is:

        .. math::

            R_x(\theta) =
            \begin{bmatrix}
            \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
            -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            theta (Union[float, int]): rotation angle
            signature (Node): node that implements the rotation gate
            **condition (int): condition of the operation
        """
        self.__check_rotation_angle(theta)
        self.__add_single_qubit_gate('rx', which_qubit, signature, angle=theta, **condition)

    def ry(self, which_qubit: int, theta: Union[float, int], signature=None, **condition) -> None:
        r"""Add a rotation gate around y-axis.

        The matrix form is:

        .. math::

            R_y(\theta) =
            \begin{bmatrix}
            \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
            \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
            \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            theta (Union[float, int]): rotation angle
            signature (Node): node that implements the rotation gate
            **condition (int): condition of the operation
        """
        self.__check_rotation_angle(theta)
        self.__add_single_qubit_gate('ry', which_qubit, signature, angle=theta, **condition)

    def rz(self, which_qubit: int, theta: Union[float, int], signature=None, **condition) -> None:
        r"""Add a rotation gate around z-axis.

        The matrix form is:

        .. math::

            R_z(\theta) =
            \begin{bmatrix}
            e^{-i\frac{\theta}{2}} & 0 \\
            0 & e^{i\frac{\theta}{2}}
            \end{bmatrix}

        Args:
            which_qubit (int): qubit index
            theta (Union[float, int]): rotation angle
            signature (Node): node that implements the rotation gate
            **condition (int): condition of the operation
        """
        self.__check_rotation_angle(theta)
        self.__add_single_qubit_gate('rz', which_qubit, signature, angle=theta, **condition)

    def u3(self, which_qubit: int, theta: Union[float, int], phi: Union[float, int], gamma: Union[float, int],
           signature=None, **condition) -> None:
        r"""Add a single qubit unitary gate.

        It has a decomposition form:

        .. math::

            \begin{align}
            U_3(\theta, \varphi, \gamma) = R_z(\varphi) R_x(\theta) R_z(\gamma) =
                \begin{bmatrix}
                    \cos\frac\theta2 & -e^{i\gamma}\sin\frac\theta2\\
                    e^{i\varphi}\sin\frac\theta2 & e^{i(\varphi+\gamma)}\cos\frac\theta2
                \end{bmatrix}
            \end{align}

        Warnings:
            Please be aware of the order of the rotation angles.

        Args:
            which_qubit (int): qubit index
            theta (Union[float, int]): the rotation angle of the Rx gate
            phi (Union[float, int]): the rotation angle of the left Rz gate
            gamma (Union[float, int]): the rotation angle of the right Rz gate
            signature (Node): node that implements the unitary gate
            **condition (int): condition of the operation
       """
        angles = [theta, phi, gamma]
        for angle in angles:
            self.__check_rotation_angle(angle)

        self.__add_single_qubit_gate('u3', which_qubit, signature, angles=angles, **condition)

    def ch(self, which_qubit: List[int], signature=None) -> None:
        r"""Add a Controlled-Hadamard gate.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CH =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \frac{1}{\sqrt{2}} & \frac{1}{\sqrt{2}} \\
                0 & 0 & \frac{1}{\sqrt{2}} & -\frac{1}{\sqrt{2}}
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (list): a list of qubit indices in the order of [control, target]
            signature (Node): node that implements the Controlled-Hadamard gate
        """
        self.__add_double_qubit_gate('ch', which_qubit, signature)

    def cx(self, which_qubit: List[int], signature=None) -> None:
        r"""Add a Controlled-X gate.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CX =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1 \\
                0 & 0 & 1 & 0
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (list): a list of qubit indices in the order of [control, target]
            signature (Node): node that implements the Controlled-X gate
        """
        self.__add_double_qubit_gate('cx', which_qubit, signature)

    def cnot(self, which_qubit: List[int], signature=None) -> None:
        r"""Add a Controlled-NOT gate.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CNOT =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1 \\
                0 & 0 & 1 & 0
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (list): a list of qubit indices in the order of [control, target]
            signature (Node): node that implements the Controlled-NOT gate
        """
        self.__add_double_qubit_gate('cx', which_qubit, signature)

    def cnot15(self, which_qubit: List[int], signature=None) -> None:
        r"""Add a Controlled-NOT gate whose measurement pattern has 15 qubits.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CNOT =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & 1 \\
                0 & 0 & 1 & 0
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (list): a list of qubit indices in the order of [control, target]
            signature (Node): node that implements the Controlled-NOT gate
        """
        self.__add_double_qubit_gate('cnot15', which_qubit, signature)

    def cy(self, which_qubit: List[int], signature=None) -> None:
        r"""Add a Controlled-Y gate.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CY =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 0 & -i \\
                0 & 0 & i & 0
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (list): a list of qubit indices in the order of [control, target]
            signature (Node): node that implements the Controlled-Y gate
        """
        self.__add_double_qubit_gate('cy', which_qubit, signature)

    def cz(self, which_qubit: List[int], signature=None) -> None:
        r"""Add a Controlled-Z gate.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CZ =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & -1
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (list): a list of qubit indices in the order of [control, target]
            signature (Node): node that implements the Controlled-Z gate
        """
        self.__add_double_qubit_gate('cz', which_qubit, signature)

    def crx(self, which_qubit: List[int], theta: Union[float, int], signature=None) -> None:
        r"""Add a Controlled-rotation gate around x-axis.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CR_x(\theta) =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} \\
                0 & 0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2}
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (List[int]): a list of qubit indices in the order of [control, target]
            theta (Union[float, int]): rotation angle
            signature (Node): node that implements the Controlled-rotation gate
        """
        self.__check_rotation_angle(theta)
        self.__add_double_qubit_gate('crx', which_qubit, signature, angle=theta)

    def cry(self, which_qubit: List[int], theta: Union[float, int], signature=None) -> None:
        r"""Add a Controlled-rotation gate around y-axis.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CR_y(\theta) =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos\frac{\theta}{2} & -\sin\frac{\theta}{2} \\
                0 & 0 & \sin\frac{\theta}{2} & \cos\frac{\theta}{2}
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (List[int]): a list of qubit indices in the order of [control, target]
            theta (Union[float, int]): rotation angle
            signature (Node): node that implements the Controlled-rotation gate
        """
        self.__check_rotation_angle(theta)
        self.__add_double_qubit_gate('cry', which_qubit, signature, angle=theta)

    def crz(self, which_qubit: List[int], theta: Union[float, int], signature=None) -> None:
        r"""Add a Controlled-rotation gate around z-axis.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CR_z(\theta) =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & e^{-i\frac{\theta}{2}} & 0 \\
                0 & 0 & 0 & e^{i\frac{\theta}{2}}
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (List[int]): a list of qubit indices in the order of [control, target]
            theta (Union[float, int]): rotation angle
            signature (Node): node that implements the Controlled-rotation gate
        """
        self.__check_rotation_angle(theta)
        self.__add_double_qubit_gate('crz', which_qubit, signature, angle=theta)

    def cu3(self, which_qubit: List[int],
            theta: Union[float, int], phi: Union[float, int], gamma: Union[float, int], signature=None) -> None:
        r"""Add a Controlled-rotation gate.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                CU_3(\theta, \varphi, \gamma) =
                \begin{bmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & \cos\frac\theta2 & -e^{i\gamma}\sin\frac\theta2 \\
                0 & 0 & e^{i\varphi}\sin\frac\theta2 & e^{i(\varphi+\gamma)}\cos\frac\theta2
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (List[int]): a list of qubit indices in the order of [control, target]
            theta (Union[float, int]): the rotation angle of the Rx gate
            phi (Union[float, int]): the rotation angle of the left Rz gate
            gamma (Union[float, int]): the rotation angle of the right Rz gate
            signature (Node): node that implements the Controlled-rotation gate
        """
        angles = [theta, phi, gamma]
        for angle in angles:
            self.__check_rotation_angle(angle)

        self.__add_double_qubit_gate('cu3', which_qubit, signature, angles=angles)

    def swap(self, which_qubit: List[int], signature=None) -> None:
        r"""Add a SWAP gate.

        Let ``which_qubit`` be ``[0, 1]``, the matrix form is:

        .. math::

            \begin{align}
                SWAP =
                \begin{bmatrix}
                    1 & 0 & 0 & 0 \\
                    0 & 0 & 1 & 0 \\
                    0 & 1 & 0 & 0 \\
                    0 & 0 & 0 & 1
                \end{bmatrix}
            \end{align}

        Args:
            which_qubit (list): qubits to swap
            signature (Node): node that implements the SWAP gate
        """
        self.__add_double_qubit_gate('cx', which_qubit, signature)
        self.__add_double_qubit_gate('cx', [which_qubit[1], which_qubit[0]], signature)
        self.__add_double_qubit_gate('cx', which_qubit, signature)

    def measure(self, which_qubit=None, signature=None) -> None:
        r"""Measure a quantum state in the computational basis.

        Measure all the qubits if no specific qubit index is given.

        Args:
            which_qubit (int): qubit index
            signature (Node): node that implements the measurement
        """
        assert which_qubit is None or isinstance(which_qubit, int), \
            f"Input {which_qubit} should be an int value."

        if which_qubit is None:  # measure all qubits in the circuit
            occupied_indices = copy.deepcopy(self.occupied_indices)
            for idx in occupied_indices:
                self.__add_single_qubit_gate('m', idx, signature, basis=[0, 'YZ', [], []])  # Z measurement

        else:
            self.__add_single_qubit_gate('m', which_qubit, signature, basis=[0, 'YZ', [], []])  # Z measurement

    def defer_measurement(self) -> None:
        r"""Defer all measurements to the end.

        Note:
            Classically conditioned gates will be transformed to controlled gates.
        """
        measure_gates = []
        for i, gate in enumerate(self._history):
            if "condition" in gate.keys():
                # Fixed single-qubit gate
                if gate['name'] in ['h', 'x', 'y', 'z']:
                    self._history[i] = {"name": 'c' + gate['name'],
                                        "which_qubit": [gate['condition'], gate['which_qubit'][0]],
                                        "signature": gate['signature']}
                # Parameterized single-qubit gate
                elif gate['name'] in ['rx', 'ry', 'rz']:
                    self._history[i] = {"name": 'c' + gate['name'],
                                        "which_qubit": [gate['condition'], gate['which_qubit'][0]],
                                        "signature": gate['signature'],
                                        "angle": gate['angle']}
                elif gate['name'] == 'u3':
                    self._history[i] = {"name": 'cu3',
                                        "which_qubit": [gate['condition'], gate['which_qubit'][0]],
                                        "signature": gate['signature'],
                                        "angles": gate['angles']}
            elif gate['name'] == 'm':
                measure_gates.append(self._history[i])

        self._history = list(filter(lambda x: x not in measure_gates, self._history))
        self._history.extend(measure_gates)

    def copy(self) -> "Circuit":
        r"""Make a copy of the current circuit.

        Returns:
            Circuit: a copy of the current circuit
        """
        new_circuit = Circuit()
        new_circuit._history = self.gate_history
        return new_circuit

    def is_equal(self, other: "Circuit") -> bool:
        r"""Check if the current circuit is equal to another circuit.

        Returns:
            bool: whether the two circuits equals
        """
        if self.width != other.width:  # check if the circuits have the same size
            return False

        for i in range(self.width):
            gates_1, gates_2 = [], []

            for gate_1, gate_2 in zip(self.gate_history, other.gate_history):
                if i in gate_1['which_qubit']:
                    gates_1.append(gate_1)
                if i in gate_2['which_qubit']:
                    gates_2.append(gate_2)

            for gate_1, gate_2 in zip(gates_1, gates_2):
                if gate_1 != gate_2:
                    if not (gate_1['name'] == gate_2['name'] == "cz" or gate_1['name'] == gate_2['name'] == "swap" and
                            set(gate_1['which_qubit']) == set(gate_2['which_qubit'])):
                        return False
        return True

    def remap_indices(self, remap=None) -> None:
        r"""Remap the indices of quantum register.

        Args:
            remap (dict, optional): remap of the indices

        Examples:
            Both the keys and the values of the remap dictionary should be of int type. Here is an example:

            >>> cir = Circuit()
            >>> cir.h(1)
            >>> cir.cnot([1, 3])
            >>> cir.cnot([3, 4])
            >>> cir.remap_indices(remap={1: 0, 3: 1, 4: 2})

            Note that the new indices should cover all indices in the current circuit, and the new indices should be
            a set of sequential integers starting from zero.

        Warnings:
            Some functionalities of MBQC module require that quantum register has sequential indices.
            We need to remap the indices before using the MBQC model and after deferring all measurements.

        Important:
            This method will directly update the current circuit. Please make a new copy of the circuit if necessary.
        """
        if remap is not None:
            assert set(remap.keys()) == set(self.occupied_indices), \
                f"The remap should contain all indices ({', '.join(str(i) for i in self.occupied_indices)}) " \
                f"of the quantum register."
            assert set(remap.values()) == set(range(self.width)), \
                f"The new indices should be a set of sequential integers starting from zero."

            # Sort by the new indices in ascending order
            remap = {value: key for key, value in zip(remap.keys(), remap.values())}
            remap = {remap[key]: key for key in sorted(remap.keys())}
        else:
            remap = {self.occupied_indices[i]: i for i in range(len(self.occupied_indices))}

        # Update the gate history
        for gate in self._history:
            new_indices = [remap[which_qubit] for which_qubit in gate["which_qubit"]]
            gate["which_qubit"] = new_indices

        print(f"\nThe quantum register indices have been remapped to (old: new): {remap}.")

    def run(self, shots: int, backend: Enum, token: Optional[str] = None) -> dict:
        r"""Run the quantum circuit with a given backend.

        Args:
            shots (int): the number of sampling
            backend (Enum): backend to run the quantum circuit
            token (str, optional): your token for QCompute backend

        Returns:
            dict: circuit results, including the circuit's name, sampling shots and sampling results
        """
        if not hasattr(backend, 'name'):
            raise ArgumentTypeError(f"{backend} has no attribute 'name'. "
                                    f"Please assign a specific backend for {backend}.")

        if backend.name in Backend.QCompute.__members__:
            results = qcompute.run_circuit(self, shots=shots, backend=backend, token=token)
        elif backend.name in Backend.MBQC.__members__:
            results = mbqc.run_circuit(self, shots=shots)
        else:
            raise ArgumentTypeError(f"Cannot find the backend {backend}.")

        cir_results = {'circuit_name': self.name, 'shots': shots, 'counts': self.sort_results(results)}
        return cir_results

    @staticmethod
    def reduce_results(results: dict, indices: List[int]) -> dict:
        r"""Reduce the circuit sampling results with specific indices.

        Args:
            results (dict): circuit sampling results to reduce
            indices (List[int]): global indices for reducing circuit results

        Returns:
            dict: reduced results
        """
        assert all(indices[i] < indices[i + 1] for i in range(len(indices) - 1)), \
            f"The indices should be in ascending order."

        reduced_results = {}
        for key, value in results.items():
            res = ""
            for i in indices:
                res += key[i]
            if res not in reduced_results.keys():
                reduced_results[res] = value
            else:
                reduced_results[res] += value

        return Circuit.sort_results(reduced_results)

    @staticmethod
    def sort_results(results: dict) -> dict:
        r"""Sort the circuit results in ascending order.

        Args:
            results (dict): circuit sampling results to sort

        Returns:
            dict: sorted results
        """
        sorted_results = {key: results[key] for key in sorted(results.keys())}
        return sorted_results

    def print_agenda(self) -> None:
        r"""Print the events scheduled for the circuit.
        """
        from qcompute_qnet.core.des import Event

        df = Event.events_to_dataframe(self.agenda)
        print(f"\nAgenda of {self.name} (unsorted):\n{df.to_string()}")

    def print_list(self) -> None:
        r"""Print the quantum circuit list.
        """
        df = pd.DataFrame(columns=["name", "which_qubit", "signature", "params"])

        for i, gate in enumerate(self.gate_history):
            gate_params = {key: gate[key]
                           for key in list(filter(lambda x: x not in ['name', 'which_qubit', 'signature'], gate))}
            circuit_info = pd.DataFrame({"name": gate['name'],
                                         "which_qubit": str(gate['which_qubit']),
                                         "signature": gate['signature'].name if gate['signature'] is not None else None,
                                         "params": str(gate_params)}, index=[f"Gate {i + 1}"])
            df = pd.concat([df, circuit_info])

        print(f"\nCircuit details:\n{df.to_string()}")

    def print_circuit(self, color: Optional[bool] = False, colors: Optional[dict] = None) -> None:
        r"""Print the circuit on the terminal.

        Args:
            color (bool, optional): whether to print a colored circuit
            colors (dict, optional): specified colors for different nodes

        Examples:

            >>> from qcompute_qnet.quantum.circuit import Circuit
            >>> cir = Circuit()
            >>> cir.x(0)
            >>> cir.y(1)
            >>> cir.h(2)
            >>> cir.z(3)
            >>> cir.s(4)
            >>> cir.rx(0, 0.5)
            >>> cir.ry(1, 0.6)
            >>> cir.rz(1, 0.4)
            >>> cir.cnot([0, 1])
            >>> cir.cz([4, 3])
            >>> cir.measure()
            >>> cir.print_circuit()
        """
        # Optional colors for gates
        gate_colors = {"red": "\033[0;31m",
                       "blue": "\033[0;34m",
                       "green": "\033[0;32m",
                       "yellow": "\033[1;33m",
                       "purple": "\033[0;35m",
                       "cyan": "\033[0;36m",
                       "brown": "\033[0;33m",
                       "dark_gray": "\033[1;30m",
                       "light_red": "\033[1;31m",
                       "light_blue": "\033[1;34m",
                       "light_green": "\033[1;32m",
                       "light_purple": "\033[1;35m",
                       "light_cyan": "\033[1;36m",
                       "light_gray": "\033[0;37m",
                       "none": "\033[0m"}

        # Count the blocks needed for the print of the whole circuit
        width = self.width
        nums_block = np.zeros(width, dtype=int)

        if color is True:
            for i, gate in enumerate(self.gate_history):
                if gate['signature'] is None:
                    print("\n\nNot all gates are assigned signatures. A monochrome circuit will be printed.")
                    color = False
                    break

            if colors is not None:
                # Sanity check for input colors
                assert isinstance(colors, dict), "'colors' should be a dictionary mapping nodes to different colors."
                for node_color in colors.values():
                    assert node_color in gate_colors, f"The {node_color} color is not supported in this version."

            else:  # assign colors for different signatures
                signatures = {gate['signature'] for gate in self.gate_history}
                assigned_colors = list(gate_colors)[:len(signatures)]
                colors = {signature: assigned_color for signature, assigned_color in zip(signatures, assigned_colors)}

        for i, gate in enumerate(self.gate_history):
            if gate['name'] in {'cx', 'cy', 'cz', 'swap', 'crx', 'cry', 'crz', 'cu3', 'ch'}:
                max_c = 0

                for p in range(min(gate['which_qubit']), max(gate['which_qubit']) + 1):
                    max_c = nums_block[p] if max_c < nums_block[p] else max_c

                for p in range(min(gate['which_qubit']), max(gate['which_qubit']) + 1):
                    nums_block[p] = max_c
                    nums_block[p] += 1

            elif 'condition' in gate.keys():
                max_c = 0

                for p in range(min(gate['which_qubit'][-1], gate['condition']),
                               max(gate['which_qubit'][-1], gate['condition'])):
                    max_c = nums_block[p] if max_c < nums_block[p] else max_c

                for p in range(min(gate['which_qubit'][-1], gate['condition']),
                               max(gate['which_qubit'][-1], gate['condition']) + 1):
                    nums_block[p] = max_c
                    nums_block[p] += 1
            else:
                for qubit in gate['which_qubit']:
                    nums_block[qubit] += 1  # count each number's frequency

        # System parameters about printing
        total_length = max(nums_block)
        total_length = total_length * 6 + 3  # the number of blocks needed
        print_list = [['-' if i % 2 == 0 else ' '] * total_length for i in range(width * 2)]  # main string for printing
        print_list_idx = np.ones(width, dtype=int) * 2  # x-axis of the current gate, begins with 2

        # System parameters about coloring
        color_loc = np.zeros(width * 2, dtype=int)  # record the location to be painted
        line_color = ['\033[0m'] * width * 2  # record the color of each line

        for i, gate in enumerate(self.gate_history):
            gate_loc = gate['which_qubit']  # operating qubit(s)

            if 'condition' in gate.keys():  # condition gates
                # Find the largest location between gate location and condition
                max_idx = 0
                condition = gate['condition']

                for p in range(min(gate_loc[0], condition), max(gate_loc[0], condition) + 1):
                    if max_idx < print_list_idx[p]:
                        max_idx = print_list_idx[p]

                for p in range(min(gate_loc[0], condition), max(gate_loc[0], condition) + 1):
                    print_list_idx[p] = max_idx

                # Print the gate
                if len(gate['name']) == 1:
                    print_list[gate_loc[0] * 2][print_list_idx[gate_loc[0]]] = gate['name'].upper()

                else:  # len(gate['name']) = 2
                    print_list[gate_loc[0] * 2][print_list_idx[gate_loc[0]]] = gate['name'][0].upper()
                    print_list[gate_loc[0] * 2][print_list_idx[gate_loc[0]] + 1] = gate['name'][1]

                # Print 'o' and '='
                print_list[condition * 2][print_list_idx[gate_loc[0]]] = 'o'

                for idx in range(total_length):
                    if print_list[condition * 2].index('M') < idx < print_list[condition * 2].index('o'):
                        if print_list[condition * 2][idx][0] == '-':
                            print_list[condition * 2][idx] = '=' + print_list[condition * 2][idx][1:]
                        if print_list[condition * 2][idx][-1] == '-':
                            print_list[condition * 2][idx] = print_list[condition * 2][idx][0:-1] + '='

                # Print '|'
                for p in range(min(gate_loc[0], condition) * 2 + 1, max(gate_loc[0], condition) * 2):
                    print_list[p][print_list_idx[gate_loc[0]]] = '|'

                # Hide the line after 'o'
                for q in range(print_list[condition * 2].index('o') + 1, total_length):
                    print_list[condition * 2][q] = ' '

                # Update the idx
                for p in range(min(gate_loc[0], condition) + 1, max(gate_loc[0], condition)):
                    print_list_idx[p] += 5

            elif len(gate_loc) == 2:  # controlled gates and 'swap'
                # Find the largest location between the control and target qubit
                max_idx = 0
                for p in range(min(gate_loc), max(gate_loc) + 1):
                    if max_idx < print_list_idx[p]:
                        max_idx = print_list_idx[p]

                for p in range(min(gate_loc), max(gate_loc) + 1):
                    print_list_idx[p] = max_idx

                # Get what to be printed and print the gate
                if gate['name'] == 'swap':
                    gate_name = 'swap'
                else:
                    gate_name = gate['name'][1:]  # remove 'c', now len(gate_name) in {1, 2}

                if len(gate_name) == 1:
                    if gate_name != 'swap':
                        print_list[gate_loc[1] * 2][print_list_idx[gate_loc[1]]] = gate_name[0].upper()
                    else:  # swap gate
                        print_list[gate_loc[1] * 2][print_list_idx[gate_loc[1]]] = 'x'

                else:  # len(gate_name) = 2
                    print_list[gate_loc[1] * 2][print_list_idx[gate_loc[1]]] = gate_name[0].upper()
                    print_list[gate_loc[1] * 2][print_list_idx[gate_loc[1]] + 1] = gate_name[1]

                # Print the control and '|'s
                print_list[gate_loc[0] * 2][print_list_idx[gate_loc[0]]] = '*'

                for p in range(min(gate_loc) * 2 + 1, max(gate_loc) * 2):
                    print_list[p][print_list_idx[gate_loc[1]]] = '|'

                # Update the idx
                for p in range(min(gate_loc) + 1, max(gate_loc)):
                    print_list_idx[p] += 5

            else:
                # Gates with name length = 1, contains{'x', 'y', 'z', 'h', 's', 't', 'i', 'm'}
                if len(gate['name']) == 1:
                    print_list[gate_loc[0] * 2][print_list_idx[gate_loc[0]]] = gate['name'].upper()

                else:
                    print_list[gate_loc[0] * 2][print_list_idx[gate_loc[0]]] = gate['name'][0].upper()
                    print_list[gate_loc[0] * 2][print_list_idx[gate_loc[0]] + 1] = gate['name'][1]

            if color:
                # Paint the gate

                # If the paint loc is in front of '|', step to its behind
                if print_list[gate_loc[-1] * 2][color_loc[gate_loc[-1] * 2] + 1] == '|':
                    print_list[gate_loc[-1] * 2][color_loc[gate_loc[-1] * 2] + 2] = \
                        gate_colors[colors[gate['signature']]] + '-'
                else:
                    print_list[gate_loc[-1] * 2][color_loc[gate_loc[-1] * 2]] = \
                        gate_colors[colors[gate['signature']]] + '-'

                # Change the line color
                line_color[gate_loc[-1] * 2] = gate_colors[colors[gate['signature']]]
                # Ensure the color of the line is same after '|'
                for loc in range(color_loc[gate_loc[-1] * 2] + 1, total_length):  # scan the element before '|'
                    if len(print_list[gate_loc[-1] * 2][loc]) > 1:
                        if print_list[gate_loc[-1] * 2][loc][-1] in {'-', '='}:
                            print_list[gate_loc[-1] * 2][loc] = line_color[gate_loc[-1] * 2] + \
                                                                print_list[gate_loc[-1] * 2][loc][-1]

                if len(gate_loc) == 2:
                    # Change the color near the control qubit
                    # ('--*--' if len(gate['name'] == 2), '-*---' if len(gate['name']) == 3)
                    lg = 1 if len(gate['name']) == 3 else 2
                    color_loc[gate_loc[0] * 2] = print_list_idx[gate_loc[0]] - lg
                    print_list[gate_loc[0] * 2][color_loc[gate_loc[0] * 2]] = \
                        gate_colors[colors[gate['signature']]] + '-'  # paint the head of the '--*--'

                    rg = len(gate['name']) + 3  # the range of blocks in the gate's color
                    print_list[gate_loc[0] * 2][color_loc[gate_loc[0] * 2] + rg] = \
                        print_list[gate_loc[0] * 2][color_loc[gate_loc[0] * 2] + rg] + line_color[gate_loc[0] * 2]

                    # Color the '|'
                    for p in range(min(gate_loc) * 2 + 1, max(gate_loc) * 2):
                        block_here = '-' if p % 2 == 0 else ' '
                        print_list[p][print_list_idx[gate_loc[-1]] - 1] = \
                            block_here + gate_colors[colors[gate['signature']]]
                        print_list[p][print_list_idx[gate_loc[-1]] + 1] = line_color[p] + block_here

                    color_loc[gate_loc[0] * 2] = print_list_idx[gate_loc[-1]] + len(gate['name']) + 2 - (
                        1 if 'c' in gate['name'] else 0
                    )

                # Update color location
                color_loc[gate_loc[-1] * 2] = print_list_idx[gate_loc[-1]] + len(gate['name']) + 2 - (
                    1 if 'c' in gate['name'] else 0
                )

                if 'condition' in gate.keys():
                    condition = gate['condition']
                    line_color[condition * 2] = gate_colors[colors[gate['signature']]]
                    # Paint the '=' and 'o'
                    q = print_list[condition * 2].index('M') + 1
                    print_list[condition * 2][q + 2] = gate_colors[colors[gate['signature']]] + '='

                    # Paint the '|'
                    for p in range(min(gate_loc[0], condition) * 2, max(gate_loc[0], condition) * 2):
                        # Change the color before '|'
                        print_list[p][print_list_idx[gate_loc[0]] - 1] = \
                            print_list[p][print_list_idx[gate_loc[0]] - 1][-1] + gate_colors[colors[gate['signature']]]

                        for r in range(q, total_length):
                            if print_list[p][r] == '|' and p % 2 == 0:
                                print_list[p][r + 1] = line_color[p] + print_list[p][r + 1][-1]

            # Update the index for printing of next gate
            for p in gate_loc:
                print_list_idx[p] += 5

        # Shorten the remaining line elements
        line_end = max(print_list_idx)
        for line in range(width * 2):
            print_list[line][line_end] = gate_colors["none"]

            for pos in range(line_end + 1, total_length):
                print_list[line][pos] = ''

        print_list = list(map(''.join, print_list))
        circuit_str = '\n'.join(print_list)

        print(f"\n{self.name}: \n\n" + circuit_str)
        if color is True:
            print("Colors:", {node.name: color for node, color in colors.items()})
